#!/usr/bin/python
from __future__ import print_function
from boConfig import BO_USER, BO_PASS, BO_STATION, BO_REGION, BOOK_KEEPING_PATH, LOGGING_PATH, LOGLEVEL
import time
import calendar
import subprocess
import json
import boDb
import random
import sys
import os
import logging
import base64

desiredExecutionTime = 1 * 60 # note: cron starts this job every 10 minutes, thus some reserve should be there
secStep = 10 * 60 # 10 minutes in seconds
bookKeepingFile = "boGetSignalsFromBlitzortung.txt"
logFile = "boGetSignalsFromBlitzortung.log"

url = "http://{}:{}@signals.blitzortung.org/Data_{}/{}/%Y/%m/%d/%H/%M.log"

new = 0
done = 0
dropped = 0

def configLogger():
    global logger
    global logFile
    format = "%(asctime)-15s %(levelname)s:%(name)s: %(message)s"
    
    path = LOGGING_PATH
    if ("" == path):
        path = "./"
    if (os.path.exists(path) and os.path.isdir(path)):
        if ('/' == path[-1]):
            filename = path + logFile
        else:
            filename = path + '/' + logFile
    else:
        print("ERROR: LOGGING_PATH = '{}' does not exist".format(LOGGING_PATH))
        sys.exit(1)

    logging.basicConfig(filename = filename, format = format)
    logger = logging.getLogger(__name__)
    logger.setLevel(LOGLEVEL)
    logger.setLevel(logging.INFO) # overrule LOGLEVEL
    boDb.configLogger()

def calcRequiredDownloads():
    """1 download to be done every 10 minutes from 3 servers
    server may not be reachable => retrials may be required
    server may also not have data => no endless retries
    file for bookkeeping
    first line: start and end time/date in human readable format
    second line: end time/date in human readable format
    following lines: downloads to be done from start to end (inclduing borders))
    succeeded lines are deleted
    missing lines for newer dates are to be added
    """
    global secStep
    global bookKeepingFile
    global new

    fileDateTimeFormat = "%Y.%m.%d %H:%M"
    firstLine = True

    urls = []
    secEnd = (int(time.time() / secStep) * secStep) - secStep # multiple of secStep, older than now
    try:
        with open(bookKeepingFile, 'rt') as f:
            for line in f:
                line = line.rstrip()
                if (line):
                    urls.append(line)
                if (firstLine):
                    firstLine = False
                    line = line.split('-')
                    tInitialStart = time.strptime(line[0].strip(), fileDateTimeFormat)
                    secInitialStart = int(calendar.timegm(tInitialStart) / secStep) * secStep
                    tInitialEnd = time.strptime(line[1].strip(), fileDateTimeFormat)
                    secInitialEnd = int(calendar.timegm(tInitialEnd) / secStep) * secStep
                    secStart = secInitialEnd + secStep
    except:
        secInitialStart = secEnd - 3 * 60 * 60 # signals are available for the previous 3 hours, thus initially start with this time range
        secStart = secInitialStart
        urls.append("")
    urls[0] = time.strftime(fileDateTimeFormat, time.gmtime(secInitialStart)) + " - " + time.strftime(fileDateTimeFormat, time.gmtime(secEnd))

    new = 0
    for sec in range(secStart, secEnd + secStep, secStep):
        t = time.gmtime(sec)
        for i in BO_REGION:
            link = url.format(BO_USER, BO_PASS, i, BO_STATION)
            link = time.strftime(link, t)
            urls.append(link)
            new += 1
    return urls

def savePendingDownloads(urls):
    global new
    global done
    global dropped
    with open(bookKeepingFile, 'wt') as f:
        for line in urls:
            f.write(line + '\n')
    logger.info("{} downloaded: {} new, {} done, {} dropped, {} pending".format(urls[0], new, done, dropped, len(urls) - 1))

def readStderr(sp, cmd):
    err = ""
    if (sp.returncode):
        err += "{} returncode: {}; ".format(cmd, sp.returncode)
    line = sp.stderr.readline()
    while line:
        err += "{}: {};".format(cmd, line)
        line = sp.stderr.readline()
    return err

def getStationSignalData(line):
    signal = {}
    j = {}
    try:
        j = json.loads(line)
    except:
        logger.error("parsing json data failed: '{}'".format(sys.exc_info()[0]))
    if ("station" in j) and (int(BO_STATION) == j["station"]):
        if ("time" in j):
            signal["time_ns"] = j["time"]
            if ("channels" in j):
                signal["channel"] = {0: {}, 1: {}, 2: {}, 3: {}}
                if ("0" in j["channels"]):
                    signal["channel"][0]["data"] = base64.b64decode(j["channels"]["0"]["data"])
                if ("1" in j["channels"]):
                    signal["channel"][1]["data"] = base64.b64decode(j["channels"]["1"]["data"])
                if ("2" in j["channels"]):
                    signal["channel"][2]["data"] = base64.b64decode(j["channels"]["2"]["data"])
                if ("3" in j["channels"]):
                    signal["channel"][3]["data"] = base64.b64decode(j["channels"]["3"]["data"])
            else:
                logger.error("no channel in the data line '{}'".format(line))
        else:
            logger.error("no time in the data line '{}'".format(line))
    else:
        logger.error("no 'station' {} in the data line '{}'".format(BO_STATION, line))
    return signal

def urlGetGrepParse(url):
    result = False

    cmd1 = "wget -qO- {}".format(url).split()
    sp1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    signalCnt = 0
    signalInsertFailed = 0
    line = sp1.stdout.readline()
    while line:
        signal = getStationSignalData(line)
        if (signal):
            if not boDb.insertSignal(signal):
                signalInsertFailed += 1
            signalCnt += 1
        line = sp1.stdout.readline()

    sp1.wait() # wait for the process to terminate in order to gte the returncode
    err = readStderr(sp1, "wget")

    urlForLogging = url.replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)
    if ("" == err):
        if (signalInsertFailed):
            logger.info("'{}' contained {} signals, insert failed for {}".format(urlForLogging, signalCnt, signalInsertFailed))
        else:
            logger.info("'{}' contained {} signals".format(urlForLogging, signalCnt))
        return True
    else:
        logger.debug("'{}' failed with error '{}'".format(urlForLogging, err))
        return False

def correlateSignalsStrokes():
    links = boDb.correlateSignalsStrokes()
    links += boDb.correlateSignalsStrokes(-999999999)
    linkCnt = 0
    linkUpdateFailed = 0
    for link in links:
        if not boDb.updateStrokeSignal(link):
            linkUpdateFailed += 1
        linkCnt += 1
    if (linkUpdateFailed):
        logger.info("correlating signals and strokes: {} links identified, update failed for {}".format(linkCnt, linkUpdateFailed))
    else:
        logger.info("correlating signals and strokes: {} links identified".format(linkCnt))

def getBookKeepingPath():
    global bookKeepingFile
    if ("" == BOOK_KEEPING_PATH):
        return
    if (os.path.exists(BOOK_KEEPING_PATH) and os.path.isdir(BOOK_KEEPING_PATH)):
        if ('/' == BOOK_KEEPING_PATH[-1]):
            bookKeepingFile = BOOK_KEEPING_PATH + bookKeepingFile
        else:
            bookKeepingFile = BOOK_KEEPING_PATH + '/' + bookKeepingFile
    else:
        logger.error("BOOK_KEEPING_PATH = '{}' does not exist".format(BOOK_KEEPING_PATH))
        sys.exit(1)

def main():
    # normal program starts here
    global done
    global dropped
    global desiredExecutionTime
    done = 0
    dropped = 0
    configLogger()
    tEnd = time.time() + desiredExecutionTime
    random.seed()
    getBookKeepingPath()
    boDb.connectMysql()
    boDb.connectDatabase()
    urlsNow = calcRequiredDownloads()
    urlsFuture = []
    urlsFuture.append(urlsNow[0]) # copy start and end date/time
    urlIdx = range(1, len(urlsNow)) # skip the first line, which is start and end date/time
    time.sleep(30) # allow the server to finalize the file - dealy due to fear to miss something not due to knowledge
    while (tEnd > time.time()):
        count = len(urlIdx)
        if (0 == count):
            break
        i = urlIdx.pop(random.randint(0, count - 1))
        url = urlsNow[i].split(',')
        if (1 == len(url)):
            url.append(0) # no error count present, start counting at 0
        result = urlGetGrepParse(url[0])
        if (True == result):
            done += 1
        else:
            url[1] = int(url[1]) + 1 # increment the error counter
            urlForLogging = url[0].replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)
            if (144 <= url[1]): # a day has 24 * 60 = 1440 minutes, best case 1 retry each 10 minutes = 144 retries a day
                # maximum number of retries is reached
                logger.error("'{}' finally failed after {} attempts".format(urlForLogging, url[1]))
                dropped += 1
            else:
                urlsFuture.append("{},{}".format(url[0], url[1])) # store the incremented error counter
                logger.info("'{}' failed {} times".format(urlForLogging, url[1]))
    for i in urlIdx:
        urlsFuture.append("{}".format(urlsNow[i]))
    savePendingDownloads(urlsFuture)
    if done:
        correlateSignalsStrokes()
    boDb.disconnectMysql()

if __name__ == "__main__":
    main()
